// SPDX-License-Identifier: LGPL-2.1-or-later
/*
 * Copyright (C) 2011      Tomas Gavenciak <gavento@ucw.cz>
 * Copyright (C) 2011-2013 Cyril Hrubis <metan@ucw.cz>
 */

 /*

   Helper macros to get/set bits given offset and lenght.

   The GP_GET_BITS() and GP_SET_BITS() works __ONLY__ on aligned data types.
   Which means that you can only pass value that is suitably aligned for it's
   type, for example passing an 32 bit integer is OK, passing a char buffer
   casted to 32 bit integer is not (unless you made sure that the start address
   is multiple of 4).

   Now the align-safe variants first gets the value from a buffer, byte by byte
   and then uses the GP_GET_BITS() or GP_SET_BITS(). The number in their name
   tells how much bytes are touched, as we need to touch minimal number of
   bytes needed.

  */

#ifndef CORE_GP_GET_SET_BITS_H
#define CORE_GP_GET_SET_BITS_H

/*
 * Helper macros to read/write parts of words
 *
 * Return (shifted) count bits at offset of value
 * Note: operates with value types same as val
 */
#define GP_GET_BITS(offset, len, val) \
	(sizeof(val) * 8 <= len ? \
	 (val)>>(offset) : \
	 ((val)>>(offset)) & (((((typeof(val))1)<<(len)) - 1)))

/*
 * Align-safe getbits
 */
#define GP_GET_BITS4_ALIGNED(offset, len, val) ({ \
	uint32_t v;                               \
	v  = ((uint8_t *)&val)[0];                \
	v |= ((uint8_t *)&val)[1]<<8;             \
	v |= ((uint8_t *)&val)[2]<<16;            \
	v |= ((uint8_t *)&val)[3]<<24;            \
                                                  \
	GP_GET_BITS(offset, len, v);              \
})

#define GP_GET_BITS3_ALIGNED(offset, len, val) ({ \
	uint32_t v;                               \
	v  = ((uint8_t *)&val)[0];                \
	v |= ((uint8_t *)&val)[1]<<8;             \
	v |= ((uint8_t *)&val)[2]<<16;            \
                                                  \
	GP_GET_BITS(offset, len, v);              \
})

#define GP_GET_BITS2_ALIGNED(offset, len, val) ({ \
	uint16_t v;                               \
	v  = ((uint8_t *)&val)[0];                \
	v |= ((uint8_t *)&val)[1]<<8;             \
                                                  \
	GP_GET_BITS(offset, len, v);              \
})

#define GP_GET_BITS1_ALIGNED(offset, len, val) ({ \
	uint8_t v;                                \
	v = ((uint8_t *)&val)[0];                 \
                                                  \
	GP_GET_BITS(offset, len, v);              \
})

/*
 * Set count bits of dest at ofset to val (shifted by offset)
 *
 * Does not check val for overflow
 * Operates on 8, 16, and 32 bit values, depending on the type of dest,
 * this should be unsigned
 *
 * GP_SET_BITS_OR anly sets (|=) the bits, assuming these are clear beforehand
 * GP_CLEAR_BITS sets the target bits to zero
 * GP_SET_BITS does both
 */
#define GP_CLEAR_BITS(offset, len, dest) \
       ((dest) &= ~(((((typeof(dest))1) << (len)) - 1) << (offset)))

#define GP_SET_BITS_OR(offset, dest, val) ((dest) |= ((val)<<(offset)))

#define GP_SET_BITS(offset, len, dest, val) do {  \
               GP_CLEAR_BITS(offset, len, dest);  \
               GP_SET_BITS_OR(offset, dest, val); \
} while (0)

/*
 * Align-safe setbits
 */
#define GP_SET_BITS1_ALIGNED(offset, len, dest, val) do { \
	uint8_t v = ((uint8_t *)dest)[0];                 \
	GP_SET_BITS(offset, len, v, val);                 \
	((uint8_t *)dest)[0] = v;                         \
} while (0)

#define GP_SET_BITS2_ALIGNED(offset, len, dest, val) do { \
	uint16_t v;                                       \
	v  = ((uint8_t *)dest)[0];                        \
	v |= ((uint8_t *)dest)[1]<<8;                     \
	                                                  \
	GP_SET_BITS(offset, len, v, val);                 \
	                                                  \
	((uint8_t *)dest)[0] = 0xff & v;                  \
	((uint8_t *)dest)[1] = 0xff & (v >> 8);           \
} while (0)

#define GP_SET_BITS3_ALIGNED(offset, len, dest, val) do { \
	uint32_t v;                                       \
	v  = ((uint8_t *)dest)[0];                        \
	v |= ((uint8_t *)dest)[1]<<8;                     \
	v |= ((uint8_t *)dest)[2]<<16;                    \
	                                                  \
	GP_SET_BITS(offset, len, v, val);                 \
	                                                  \
	((uint8_t *)dest)[0] = 0xff & v;                  \
	((uint8_t *)dest)[1] = 0xff & (v >> 8);           \
	((uint8_t *)dest)[2] = 0xff & (v >> 16);          \
} while (0)

#define GP_SET_BITS4_ALIGNED(offset, len, dest, val) do { \
	uint32_t v;                                       \
	v  = ((uint8_t *)dest)[0];                        \
	v |= ((uint8_t *)dest)[2]<<16;                    \
	v |= ((uint8_t *)dest)[3]<<24;                    \
	                                                  \
	GP_SET_BITS(offset, len, v, val);                 \
	                                                  \
	((uint8_t *)dest)[0] = 0xff & v;                  \
	((uint8_t *)dest)[1] = 0xff & (v >> 8);           \
	((uint8_t *)dest)[2] = 0xff & (v >> 16);          \
	((uint8_t *)dest)[3] = 0xff & (v >> 24);          \
} while (0)


#endif /* CORE_GP_GET_SET_BITS_H */
