// SPDX-License-Identifier: LGPL-2.1-or-later
/*
 * Copyright (C) 2009-2010 Jiri "BlueBear" Dluhos
 *                         <jiri.bluebear.dluhos@gmail.com>
 *
 * Copyright (C) 2009-2012 Cyril Hrubis <metan@ucw.cz>
 *
 * Copyright (C) 2011      Tomas Gavenciak <gavento@ucw.cz>
 */

#ifndef CORE_GP_TRANSFORM_H
#define CORE_GP_TRANSFORM_H

#include "core/gp_common.h"
#include <core/gp_fixed_point.h>

/*
 * Flip a coordinate within pixmap according to pixmap transformation.
 */
#define GP_TRANSFORM_X(pixmap, x) do {   \
	if ((pixmap)->x_swap)            \
		x = (pixmap)->w - x - 1; \
} while (0)

#define GP_TRANSFORM_Y(pixmap, y) do {   \
	if ((pixmap)->y_swap)            \
		y = (pixmap)->h - y - 1; \
} while (0)

/*
 * Swap coordinates (axes) according to pixmap transformation.
 */
#define GP_TRANSFORM_SWAP(pixmap, x, y) do { \
	if ((pixmap)->axes_swap)             \
		GP_SWAP(x, y);                \
} while (0)

/*
 * Transform "user"-coordinates to "real"-coordinates according to pixmap
 * transformation.
 */
#define GP_TRANSFORM_POINT(pixmap, x, y) do { \
	GP_TRANSFORM_SWAP(pixmap, x, y);      \
	GP_TRANSFORM_X(pixmap, x);            \
	GP_TRANSFORM_Y(pixmap, y);            \
} while (0)

#define GP_TRANSFORM_POINT_FP(pixmap, x, y) do { \
	GP_TRANSFORM_SWAP(pixmap, x, y);         \
	GP_TRANSFORM_X_FP(pixmap, x);            \
	GP_TRANSFORM_Y_FP(pixmap, y);            \
} while (0)

/*
 * Transform "user"-coordinates to "real"-coordinates of a rectangle corner
 * according to pixmap transformation. Corner with min-coordinates is
 * transformed to (different) corner with min-coordinates etc.
 * Arguments x, y, w, h are modified.
 */
#define GP_TRANSFORM_RECT(pixmap, x, y, w, h) do { \
	GP_TRANSFORM_SWAP(pixmap, x, y);           \
	GP_TRANSFORM_SWAP(pixmap, w, h);           \
	                                            \
	if ((pixmap)->x_swap)                      \
		x = (pixmap)->w - x - w;           \
	                                            \
	if ((pixmap)->y_swap)                      \
		y = (pixmap)->h - y - h;           \
} while (0)

#define GP_TRANSFORM_RECT_FP(pixmap, x, y, w, h) do {    \
	GP_TRANSFORM_SWAP(pixmap, x, y);                 \
	GP_TRANSFORM_SWAP(pixmap, w, h);                 \
	                                                  \
	if ((pixmap)->x_swap)                            \
		x = GP_FP_FROM_INT((pixmap)->w) - x - w; \
	                                                  \
	if ((pixmap)->y_swap)                            \
		y = GP_FP_FROM_INT((pixmap)->h) - y - h; \
} while (0)

/*
 * Transform "user"-coordinates to "real"-coordinates for a blit
 * called as gp_blit(c1, x1, y1, w, h, c2, x2, y2).
 * All x1, y1, x2, y2, w, h are adjusted.
 */
#define GP_TRANSFORM_BLIT(c1, x1, y1, w, h, c2, x2, y2) do { \
	GP_TRANSFORM_RECT(c1, x1, y1, w, h);            \
	int w2 = w, h2 = h;                             \
	GP_TRANSFORM_SWAP(c2, w2, h2);                  \
	GP_TRANSFORM_RECT(c2, x2, y2, w2, h2);          \
} while (0)

/*
 * Inverse transformation to GP_TRANSFORM_POINT.
 * Use for translating mouse pointer coordinates to coordinates on pixmap.
 */
#define GP_RETRANSFORM_POINT(pixmap, x, y) do { \
	GP_TRANSFORM_X(pixmap, x);              \
	GP_TRANSFORM_Y(pixmap, y);              \
	GP_TRANSFORM_SWAP(pixmap, x, y);        \
} while (0)

#endif /* CORE_GP_TRANSFORM_H */
